#include <vector>
#include <stdlib.h> /* rand() */
#include <sys/time.h>
#include <set>
#include <map>

#include <boost/random/mersenne_twister.hpp>
#include <boost/random/uniform_int.hpp>
#include <boost/random/variate_generator.hpp>


#define BISQBOT_INTERNAL
#include "bisqbotdefs.hh"
#include "bisqbot.hh"
extern "C" unsigned char RAM[0x800];
extern "C" unsigned char PALRAM[0x20];
extern "C" void FCEU_DispMessage(const char *format, ...)
  __attribute__((format(printf,1,2)));

extern "C" unsigned char RefrRAM[0x800];
unsigned char RefrRAM[0x800];

static unsigned NextChange=0;

static boost::mt19937 rng;

#include "rockman1sound.hh"
static class Rockman1Sound Rockman1Sound;
#include "rockman1sound.cc"

struct BisqBotStateData
{
    /* Put here all variables that you want to be saved
     * with Savestates when the core does rerecording.
     * Reset() is only called when seeking back to frame 0,
     * which is not always the case if you use 2 or more steps.
     */
    int Goal;
    double BonusX, BonusY;
    unsigned NumShoots;
    unsigned MaxShoots;
    unsigned EventFrame;
    bool Region;
};
static class BisqBotState: public BisqBotStateBase, public BisqBotStateData
{
public:
    virtual ~BisqBotState() { }
    
    BisqBotState()
    {
        Reset();
    }
    void Reset()
    {
        Goal = 0;
        NextChange = 0;
        NumShoots = 0;
        BonusX = 0;
        BonusY = 0;
        MaxShoots = 1;
        Region = false;
        EventFrame = 0;
    }
    
    void operator= (const BisqBotState& b)
    {
        (BisqBotStateData&)*this = b;
    }

public:
} state;

const BisqBotStateBase* BisqBotNewSaveState()
{
    return new BisqBotState(state);
}

void BisqBotLoadState(const BisqBotStateBase* b)
{
    state = *(BisqBotState*)b;
}

extern "C" {
void FCEUI_GetIVectors(unsigned short *reset, unsigned short *irq, unsigned short *nmi);
unsigned SavedHScroll,SavedVScroll;
}

void BisqBotMonitorCPU
  (unsigned short PC,
   unsigned char A,
   unsigned char X,
   unsigned char Y)
{
    static unsigned short irq=0, nmi=0, res=0;
    if(!irq) { FCEUI_GetIVectors(&res,&irq,&nmi); }

#if 0
    #if 0
    static std::map<unsigned, std::set<unsigned> > jptpc;
    static unsigned prevpc;
    if(prevpc == 0x83EE
    || prevpc == 0x8276
    || prevpc == 0x8803
    || prevpc == 0x8EBD
    || prevpc == 0xB821
    || prevpc == 0xF243)
    {
        if(jptpc[prevpc].find(PC) == jptpc[prevpc].end())
        {
            FCEU_DispMessage("New PC $%X in table $%X\n", PC, prevpc);
            jptpc[prevpc].insert(PC);
        }
    }
    prevpc = PC;
    #endif
    if(PC == 0xC563) { state.Goal += 1; FCEU_DispMessage("Got KEY\n"); }
    if(PC == 0xC587) { state.Goal += 2; FCEU_DispMessage("Got DOOR\n"); }
#endif

#if 0
static unsigned prevpc;
unsigned addr=PC;
    if(0
|| (addr >= 0x80F0 && addr <= 0x80F2)
|| (addr >= 0x81D8 && addr <= 0x81D8)
|| (addr >= 0x81DA && addr <= 0x81DA)
)
    {
        FCEU_DispMessage("PC=%X A=%02X X=%02X Y=%02X prevPC=%04X\n", PC,A,X,Y, prevpc);
        exit(1);
    }
    prevpc=PC;
#endif

    
#if 0 //ROCKMAN1
    static bool sound=false;
    
    if(PC == 0xD54C) sound=true;
    if(PC == 0xD56E) sound=false;
    //if(sound && PC == 0x9000) Rockman1Sound.Update();
    //if(sound && PC == 0x9003) Rockman1Sound.Command(A,Y);
    
    if(!sound && PC == 0xA238)
    {
        FCEU_DispMessage("Damage from enemy %02X (%u points)\n", Y,A);
        state.Goal = -5;
    }
    if(!sound && PC == 0xBEDF)
    {
        FCEU_DispMessage("Enemy %02X hit with weapon %u (%d points left) - rndseed=%02X\n",
            (unsigned)Y, (unsigned)RAM[0x5F], (signed char)A, (unsigned)RAM[0x46]);
        if(A==0)
            switch(RAM[0x46])
            {
                case 0x5F: case 0x60:
                case 0xC3: case 0xC4:
                    break;
                default:
                    state.Goal = -5;
            }
        if(Y == 0x3A) state.Goal = 2; // bubblebot
    }
    if(!sound
    && (PC == 0x9FE8)) // horiz scrolling
    {
        FCEU_DispMessage("Scrolling routine at PC=%04X A=%02X X=%02X Y=%02X\n", PC, A,X,Y);
        state.Goal = 2;
    }
    if(!sound
    && (PC == 0x925D)) // scroll prev room
    {
        FCEU_DispMessage("Scroll previous room at PC=%04X A=%02X X=%02X Y=%02X\n", PC, A,X,Y);
    }
    if(!sound
    && (PC == 0x92A6)) // scroll next room
    {
        FCEU_DispMessage("Scroll next room at PC=%04X A=%02X X=%02X Y=%02X\n", PC, A,X,Y);
        state.Goal = 2;
    }
    if(PC == 0xC912)
    {
        state.Goal += 2;
        FCEU_DispMessage("Boss %u damaged with weapon %u (%d points left)\n", RAM[0xAC], RAM[0x5F], (signed char)A);
    }
    if(PC == 0xC833)
    {
        FCEU_DispMessage("Acquire item 0x%02X\n", A);
        
        //if(RAM[0x46] >= 0x30) state.Goal += 2; else
        //if(A == 0x3F) state.Goal += 2; else
        state.Goal = -5;
    }
    if(!sound
    && (PC == 0xBF25))
    {
        unsigned objnum = RAM[0x2F];
        state.BonusX = RAM[0x460+objnum]*256 + (unsigned short)(RAM[0x480+objnum]*256 + RAM[0x4A0+objnum])/ 256.0;
        state.BonusY = (unsigned short)(RAM[0x600+objnum]*256 + RAM[0x620+objnum])/ 256.0;

        if(Y == 0x3B)
            FCEU_DispMessage("Create no bonus at %.3f,%.3f\n", state.BonusX,state.BonusY);
        else
            FCEU_DispMessage("Create bonus 0x%02X at %.3f,%.3f\n", (unsigned)Y, state.BonusX,state.BonusY);
        if(Y == 0x3F) state.Goal += 2; else state.Goal = -5;
        /**/
        if(state.BonusX >= 6941.9) state.Goal = -5;
        if(state.BonusX >= 6888.0 && state.BonusX < 6905.0) state.Goal = -5;
        /**/
        //if(state.BonusX < 6872.0) state.Goal = -5;
    }
    
    if(PC == 0xD4B1 && !BotFrontActive()) FCEU_DispMessage("LAG detected\n");
#endif
    if(PC == 0xD4AD) std::memcpy(RefrRAM, RAM, 0x800);

#if 1 //ROCKMAN2
    // 90C9 = DoScrolling
    if(true
    && (PC == 0x9185)) // vert scrolling
    {
        FCEU_DispMessage("VertScroll at PC=%04X A=%02X X=%02X Y=%02X\n", PC, A,X,Y);
        state.Goal += 0;
    }
    if(true
    && (PC == 0x90D2)) // horiz scrolling
    {
        FCEU_DispMessage("HorizScroll at PC=%04X A=%02X X=%02X Y=%02X\n", PC, A,X,Y);
        state.Goal += 2;
    }
    if(true
    && (PC == 0x8F39)) // next-scrolling
    {
        FCEU_DispMessage("ScrollNextRoom at PC=%04X A=%02X X=%02X Y=%02X -- $37=%u, Rockman Y=%.2f speed %.2f, item1 %02X\n",
            PC, A,X,Y, RAM[0x37],
            (unsigned short)(RAM[0x4A0]*256 + RAM[0x4C0])/256.0,
            (signed short)(RAM[0x640]*256 + RAM[0x660])/256.0,
            RAM[0xA4]);
        if(RAM[0x37] & 1) state.Goal = -5; else state.Goal += 2;
        // vert = bad, horiz = good
    }
    if(PC == 0x80AB)
    {
        FCEU_DispMessage("Boss %u damaged with weapon %u (%d points left)\n", RAM[0xB3], RAM[0xB3], (signed char)A);
    }
    
    if(!BotFrontActive())
    {
        if(/*PC == 0xCFFA // megaman2
        || */PC == 0xCFF7 // rockman2
          ) FCEU_DispMessage("LAG detected at PC=%04X\n", PC);
    }
    
    /*
    static int counter=0;
    if(++counter>=50000) {counter=0;
    fprintf(stderr, "Object types: ");
    for(unsigned a=0; a<32; ++a)
        fprintf(stderr, " %u", RAM[0x400 + a]);
    fprintf(stderr, "\n");}*/
    
    /*if(PC == nmi)
    {
        FCEU_DispMessage("Y=%.2f speed=%.2f\n",
            (unsigned short)(RAM[0x4A0]*256 + RAM[0x4C0])/256.0,
            (signed short)(RAM[0x640]*256 + RAM[0x660])/256.0);
    }*/
#endif

#if 0
    if(false && PC == 0x92B6)
    {
        FCEU_DispMessage("stage=%02X A=%02X CurScreen=%02X\n", RAM[0x31], A, RAM[0x460]);
    }
    if(false && PC == 0xCD8A)
    {
        FCEU_DispMessage("Y=%u 0C=$%02X 0D=$%02X 0E=$%02X\n", Y,RAM[12],RAM[13],RAM[14]);
        fprintf(stderr, "#Actives=%u [%f] (", RAM[0x720], RAM[0x8E]/6.0);
        for(unsigned a=0; a<RAM[0x720]; ++a)
        {
            fprintf(stderr, "%u,%u,%u,%u,%u,%u, ",
                RAM[0x721+a*6],
                RAM[0x722+a*6],
                RAM[0x723+a*6],
                RAM[0x724+a*6],
                RAM[0x725+a*6],
                RAM[0x726+a*6]);
        }
        fprintf(stderr, ")\n");
    }
    //if(PC == 0xCDDE) fprintf(stderr, "CDDE y=%u\n", Y);
#endif
}

void BisqBotReset()
{
    struct timeval tv;
    gettimeofday(&tv,NULL);
    rng.seed(tv.tv_usec + tv.tv_sec*1000000ULL);
    
    state.Reset();
}

template<unsigned mod>
static inline unsigned GetRnd()
{
    static boost::uniform_int<int> g(0, mod-1);
    return g(rng);
}
static inline unsigned GetRnd(unsigned mod)
{
    return boost::uniform_int<int> (0,mod-1) (rng);
}

unsigned char BisqBotInput(unsigned FrameNo)
{
    static const unsigned char InputTable[] =
    {
        K_R,
        K_R,
        K_R | K_A,
        K_R | K_A,
        K_R,
        K_R,
        K_R | K_A,
        K_R | K_A,
        K_L,
        K_L | K_A,
        0
    };
    static unsigned LastInput = 0;
    static unsigned A_State   = 0;
    static unsigned B_State   = 0;
    
    if(!NextChange) 
    {
        LastInput = InputTable[ GetRnd<sizeof(InputTable)>() ];
        /*if(FrameNo < 6)
            NextChange = 0;
        else*/
            NextChange = GetRnd<10>();
    }
    else --NextChange;

#if 0
    double RockX  =
        RAM[0x460]*256 + (unsigned short)(RAM[0x480]*256 + RAM[0x4A0])/ 256.0;
    double RockY  = (unsigned short)(RAM[0x600]*256 + RAM[0x620])/ 256.0;
    double RockYS = (signed short)(RAM[0x680]*256 + RAM[0x660])/ 256.0;
    double NextY = RockY-RockYS;
#endif

 //   if(RockX >= 6756 && RockX < 6768) LastInput |= K_A;
 //   if(RockX >= 6740 && RockX < 6756) LastInput &= ~K_A;
    //if(FrameNo < 10) { LastInput &= ~K_L; LastInput |= K_R; }
/*
    if(RockY >= 1.5) LastInput = K_R;
*/
    
    /*
    if(B_State && FrameNo)
    {
        double RockX = (unsigned short)(RAM[0x480]*256 + RAM[0x4A0])/ 256.0;
        fprintf(stderr,
            "[frame %u]I think the beam went at %u... X=%.1f screen=%u\n",
                FrameNo, RAM[0x5F5], RockX, RAM[0x460]);
    }
    
    B_State = (NextY >= 112 && NextY <= 114) ? K_B : 0;
    if(B_State) fprintf(stderr, "Try beam at RockY=%.1f (%+.1f)\n", RockY, RockYS);
    if(RockY >= 104 && RockY <= 112 && NextY > 113)
        fprintf(stderr, "Rats, this try failed Y=%.1f (%+.1f)\n", RockY, RockYS);
    */
    
    /*
    if(FrameNo < 7) if(GetRnd<3>() == 0) Input |= K_SE;
    if(FrameNo >= 12 && FrameNo < 45) if(GetRnd<20>() == 0) Input |= K_ST;
    */
    
    //if(FrameNo < 16) A_State = K_A; else A_State = 0;
    //if(FrameNo < 43) B_State = K_B; else B_State = 0;
    
    bool Region = false;//RockY >= 78 && RockY < 107;
    if(state.NumShoots < state.MaxShoots && GetRnd<15>() == 0
#if 0
    && (!Region || FrameNo >= 27)
#endif
      )
        { ++state.NumShoots; B_State = K_B;
          if(Region) state.Region=true;
          state.EventFrame = FrameNo;
        }
    else
        B_State = 0;
    
    //if(B_State == K_B) { LastInput &= ~K_L; LastInput |= K_R; }
    //if(FrameNo >= 8 && FrameNo < 15) LastInput |= K_A;
    
    //if(GetRnd<8>() != 0) A_State = K_A; else A_State = 0;
    //if(GetRnd<20>() == 0) Input |= K_U;
    //if(GetRnd<10>() == 0) Input |= K_ST;
    //if(GetRnd<100>() == 0) Input |= K_SE;
    //if(RockY < 1) A_State = 0;

    unsigned char Input = 0x00;
    Input |= LastInput;
    Input |= B_State;
    Input |= A_State;
    
    return Input;
}

unsigned long long BisqBotEvaluate(unsigned FrameNo)
{
    /*
    #define OkBeam(y) ((y)==113)
    
    state.Goal=1;
    if(OkBeam(RAM[0x5F5]) || OkBeam(RAM[0x5F6])
    || OkBeam(RAM[0x5F7]) || OkBeam(RAM[0x5F8])
    || OkBeam(RAM[0x5F9]))
    {
        if(RAM[0x460] == 5) // screen 5
            state.Goal = 2;
    }*/
    
    /*
    // Fail if screen 5 isn't reached in time
    if(FrameNo >= 40 && RAM[0x460] != 5) { state.Goal=-1; return 0; }
    // Fail if Rockman falls off the beam too soon
    if(FrameNo <= 32 && (signed char)RAM[0x600] > 5) { state.Goal=-1; return 0; }
    if(state.Goal < 0) return 0;
    */
    //if(FrameNo==60) state.Goal = 2;

#if 0 /* ROCKMAN 1 */
    double RockY  = (unsigned short)(RAM[0x600]*256 + RAM[0x620])/ 256.0;
    double RockX  =
        RAM[0x460]*256 + (unsigned short)(RAM[0x480]*256 + RAM[0x4A0])/ 256.0;
    double RockYS = (signed short)(RAM[0x680]*256 + RAM[0x660])/ 256.0;
#endif
#if 1 /* ROCKMAN 2 */
    double RockY  = (unsigned short)(RAM[0x4A0]*256 + RAM[0x4C0])/ 256.0;
    double RockX  =
        RAM[0x440]*256 + (unsigned short)(RAM[0x460]*256 + RAM[0x480])/ 256.0;
    double RockYS = (signed short)(RAM[0x620]*256 + RAM[0x640])/ 256.0;
#endif

    //if(state.Goal < 1 && RockY >= 117.0) { state.Goal = -5; }
    /*
    if(RockX >= 6848 && RockX < 6868 && RockY + RockYS >= 99) state.Goal = -5;
    */

    if(state.Goal > 0)
    {
#if 0 /* Shot distances in Rockman 1 */
        double NearestDistance = 256;
        double NearestShot = RockX + 256;
        
        for(unsigned n=2; n<5; ++n)
        {
            if(RAM[0x600+n] == 0xF8) continue; // 
            double ShotX  =
                RAM[0x460+n]*256 + (unsigned short)(RAM[0x480+n]*256 + RAM[0x4A0+n])/ 256.0;
            double Distance = ShotX - RockX;
            if(Distance < 0) Distance = -Distance;
            
            if(Distance < NearestDistance)
                { NearestDistance = Distance; NearestShot = ShotX; }
        }
#endif

        return 10000000
            - FrameNo*256*130
         //   - (unsigned)((RockX-6700)      *256*10)
         //   + (unsigned)((RockY)    *256*20)
         //   - state.EventFrame*256*10
         //   + (unsigned)(state.BonusX-6700)
         //   + (unsigned)((NearestShot-6700)*256*30)
         ;
        
        //return (10000-FrameNo*3) + RAM[0xA4]*20;
        /*
        double RockY  = (unsigned short)(RAM[0x4A0]*256 + RAM[0x4C0])/ 256.0;
        double YSpeed = (  signed short)(RAM[0x640]*256 + RAM[0x660])/-256.0;
        
        return (10000-FrameNo*3) + (unsigned)(RockY + YSpeed*3); // ok
        */
    }

    return 0;//((RAM[0x460]-4)*256+RAM[0x480]); // not done
}

enum BisqBotInstruction BisqBotTick(unsigned FrameNo, unsigned Step)
{
    if(state.Goal < 0) return BisqBotRestartStep;
#if 1
    if(state.Goal < 2 && FrameNo >= 33 && !state.Region) return BisqBotRestart;
#endif

    if(Step == 0 && state.Goal == 1)
    {
        state.MaxShoots = state.NumShoots+1;
        return BisqBotNextStep;
    }
    if(state.Goal >= 2)
    {
        double RockY  = (unsigned short)(RAM[0x4A0]*256 + RAM[0x4C0])/ 256.0;
        BotFrontMsg("RockY=%.2f, BonusX=%.3f, Goal=2\n", RockY, state.BonusX);
        return BisqBotAccept;
    }
    
    // Max item1s used: 5
    //if(RAM[0xA4] < 0x1C - 2*5) return BisqBotRestart;
    
    return BisqBotContinue;
}
